﻿using System;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.HID
{
    public class Device : IDevice
    {
        public event InsertedEventHandler Inserted;
        public event RemovedEventHandler Removed;

        private readonly string _description;
        private readonly string _devicePath;
        private readonly HidDeviceAttributes _deviceAttributes;

        private readonly HidDeviceCapabilities _deviceCapabilities;
        private DeviceMode _deviceReadMode = DeviceMode.NonOverlapped;
        private DeviceMode _deviceWriteMode = DeviceMode.NonOverlapped;
        private ShareMode _deviceShareMode = ShareMode.ShareRead | ShareMode.ShareWrite;

        private readonly DeviceEventMonitor _deviceEventMonitor;

        private bool _monitorDeviceEvents;
        protected delegate DeviceData ReadDelegate(int timeout);
        protected delegate HidReport ReadReportDelegate(int timeout);
        private delegate bool WriteDelegate(byte[] data, int timeout);
        private delegate bool WriteReportDelegate(HidReport report, int timeout);

        internal Device(string devicePath, string description = null)
        {
            _deviceEventMonitor = new DeviceEventMonitor(this);
            _deviceEventMonitor.Inserted += DeviceEventMonitorInserted;
            _deviceEventMonitor.Removed += DeviceEventMonitorRemoved;

            _devicePath = devicePath;
            _description = description;

            try
            {
                var hidHandle = OpenDeviceIO(_devicePath, NativeMethods.ACCESS_NONE);

                _deviceAttributes = GetDeviceAttributes(hidHandle);
                _deviceCapabilities = GetDeviceCapabilities(hidHandle);

                CloseDeviceIO(hidHandle);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error querying HID device '{0}'.", devicePath), exception);
            }
        }

        public IntPtr ReadHandle { get; private set; }
        public IntPtr WriteHandle { get; private set; }
        public bool IsOpen { get; private set; }
        public bool IsConnected { get { return Devices.IsConnected(_devicePath); } }
        public string Description { get { return _description; } }
        public HidDeviceCapabilities Capabilities { get { return _deviceCapabilities; } }
        public HidDeviceAttributes Attributes { get { return _deviceAttributes; } }
        public string DevicePath { get { return _devicePath; } }

        public bool MonitorDeviceEvents
        {
            get { return _monitorDeviceEvents; }
            set
            {
                if (value & _monitorDeviceEvents == false)
                {
                    _deviceEventMonitor.Init();
                }

                _monitorDeviceEvents = value;
            }
        }

        public override string ToString()
        {
            return string.Format("VendorID={0}, ProductID={1}, Version={2}, DevicePath={3}",
                                _deviceAttributes.VendorHexId,
                                _deviceAttributes.ProductHexId,
                                _deviceAttributes.Version,
                                _devicePath);
        }

        public void OpenDevice()
        {
            OpenDevice(DeviceMode.NonOverlapped, DeviceMode.NonOverlapped, ShareMode.ShareRead | ShareMode.ShareWrite);
        }

        public void OpenDevice(DeviceMode readMode, DeviceMode writeMode, ShareMode shareMode)
        {
            if (IsOpen)
            {
                return;
            }

            _deviceReadMode = readMode;
            _deviceWriteMode = writeMode;
            _deviceShareMode = shareMode;

            try
            {
                ReadHandle = OpenDeviceIO(_devicePath, readMode, NativeMethods.GENERIC_READ, shareMode);
                WriteHandle = OpenDeviceIO(_devicePath, writeMode, NativeMethods.GENERIC_WRITE, shareMode);
            }
            catch (Exception exception)
            {
                IsOpen = false;
                throw new Exception("EonaCat HID: Error opening HID device.", exception);
            }

            IsOpen = (ReadHandle.ToInt32() != NativeMethods.INVALID_HANDLE_VALUE &&
                        WriteHandle.ToInt32() != NativeMethods.INVALID_HANDLE_VALUE);
        }


        public void CloseDevice()
        {
            if (!IsOpen)
            {
                return;
            }

            CloseDeviceIO(ReadHandle);
            CloseDeviceIO(WriteHandle);
            IsOpen = false;
        }

        public DeviceData Read()
        {
            return Read(0);
        }

        public DeviceData Read(int timeout)
        {
            if (IsConnected)
            {
                if (IsOpen == false)
                {
                    OpenDevice(_deviceReadMode, _deviceWriteMode, _deviceShareMode);
                }

                try
                {
                    return ReadData(timeout);
                }
                catch
                {
                    return new DeviceData(DeviceData.ReadStatus.ReadError);
                }

            }
            return new DeviceData(DeviceData.ReadStatus.NotConnected);
        }

        public void Read(ReadCallback callback)
        {
            Read(callback, 0);
        }

        public void Read(ReadCallback callback, int timeout)
        {
            var readDelegate = new ReadDelegate(Read);
            var asyncState = new StateAsync(readDelegate, callback);
            readDelegate.BeginInvoke(timeout, EndRead, asyncState);
        }

        public async Task<DeviceData> ReadAsync(int timeout = 0)
        {
            var readDelegate = new ReadDelegate(Read);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<DeviceData>.Factory.StartNew(() => readDelegate.Invoke(timeout));
#else
            return await Task<DeviceData>.Factory.FromAsync(readDelegate.BeginInvoke, readDelegate.EndInvoke, timeout, null);
#endif
        }

        public HidReport ReadReport(int timeout = 0)
        {
            return new HidReport(Capabilities.InputReportByteLength, Read(timeout));
        }

        public void ReadReport(ReadReportCallback callback, int timeout = 0)
        {
            var readReportDelegate = new ReadReportDelegate(ReadReport);
            var asyncState = new StateAsync(readReportDelegate, callback);
            readReportDelegate.BeginInvoke(timeout, EndReadReport, asyncState);
        }

        public async Task<HidReport> ReadReportAsync(int timeout = 0)
        {
            var readReportDelegate = new ReadReportDelegate(ReadReport);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<HidReport>.Factory.StartNew(() => readReportDelegate.Invoke(timeout));
#else
            return await Task<HidReport>.Factory.FromAsync(readReportDelegate.BeginInvoke, readReportDelegate.EndInvoke, timeout, null);
#endif
        }

        public HidReport ReadReportSync(byte reportId)
        {
            byte[] cmdBuffer = new byte[Capabilities.InputReportByteLength];
            cmdBuffer[0] = reportId;
            bool bSuccess = NativeMethods.HidD_GetInputReport(ReadHandle, cmdBuffer, cmdBuffer.Length);
            DeviceData deviceData = new DeviceData(cmdBuffer, bSuccess ? DeviceData.ReadStatus.Success : DeviceData.ReadStatus.NoDataRead);
            return new HidReport(Capabilities.InputReportByteLength, deviceData);
        }

        public bool ReadFeatureData(out byte[] data, byte reportId = 0)
        {
            if (_deviceCapabilities.FeatureReportByteLength <= 0)
            {
                data = new byte[0];
                return false;
            }

            data = new byte[_deviceCapabilities.FeatureReportByteLength];

            var buffer = CreateFeatureOutputBuffer();
            buffer[0] = reportId;

            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = ReadHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_devicePath, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetFeature(hidHandle, buffer, buffer.Length);

                if (success)
                {
                    Array.Copy(buffer, 0, data, 0, Math.Min(data.Length, _deviceCapabilities.FeatureReportByteLength));
                }
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing HID device '{0}'.", _devicePath), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != ReadHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }

            return success;
        }

        public string ReadProduct()
        {
            if (ReadProduct(out byte[] data))
            {
                return System.Text.Encoding.ASCII.GetString(data).Replace("\0", string.Empty);
            }
            return string.Empty;
        }

        public bool ReadProduct(out byte[] data)
        {
            data = new byte[254];
            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = ReadHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_devicePath, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetProductString(hidHandle, data, data.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing HID device '{0}'.", _devicePath), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != ReadHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }

            return success;
        }

        public string ReadManufacturer()
        {
            if (ReadManufacturer(out byte[] data))
            {
                return System.Text.Encoding.ASCII.GetString(data).Replace("\0", string.Empty);
            }
            return string.Empty;
        }

        public bool ReadManufacturer(out byte[] data)
        {
            data = new byte[254];
            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = ReadHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_devicePath, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetManufacturerString(hidHandle, data, data.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing HID device '{0}'.", _devicePath), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != ReadHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }

            return success;
        }

        public string ReadSerialNumber()
        {
            if (ReadSerialNumber(out byte[] data))
            {
                return System.Text.Encoding.ASCII.GetString(data).Replace("\0", string.Empty);
            }
            return string.Empty;
        }

        public bool ReadSerialNumber(out byte[] data)
        {
            data = new byte[254];
            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = ReadHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_devicePath, NativeMethods.ACCESS_NONE);
                }

                success = NativeMethods.HidD_GetSerialNumberString(hidHandle, data, data.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing HID device '{0}'.", _devicePath), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != ReadHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }

            return success;
        }

        public bool Write(byte[] data)
        {
            return Write(data, 0);
        }

        public bool Write(byte[] data, int timeout)
        {
            if (IsConnected)
            {
                if (IsOpen == false)
                {
                    OpenDevice(_deviceReadMode, _deviceWriteMode, _deviceShareMode);
                }

                try
                {
                    return WriteData(data, timeout);
                }
                catch
                {
                    return false;
                }
            }
            return false;
        }

        public void Write(byte[] data, WriteCallback callback)
        {
            Write(data, callback, 0);
        }

        public void Write(byte[] data, WriteCallback callback, int timeout)
        {
            var writeDelegate = new WriteDelegate(Write);
            var asyncState = new StateAsync(writeDelegate, callback);
            writeDelegate.BeginInvoke(data, timeout, EndWrite, asyncState);
        }

        public async Task<bool> WriteAsync(byte[] data, int timeout = 0)
        {
            var writeDelegate = new WriteDelegate(Write);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<bool>.Factory.StartNew(() => writeDelegate.Invoke(data, timeout));
#else
            return await Task<bool>.Factory.FromAsync(writeDelegate.BeginInvoke, writeDelegate.EndInvoke, data, timeout, null);
#endif
        }

        public bool WriteReport(HidReport report)
        {
            return WriteReport(report, 0);
        }

        public bool WriteReport(HidReport report, int timeout)
        {
            return Write(report.GetBytes(), timeout);
        }

        public void WriteReport(HidReport report, WriteCallback callback)
        {
            WriteReport(report, callback, 0);
        }

        public void WriteReport(HidReport report, WriteCallback callback, int timeout)
        {
            var writeReportDelegate = new WriteReportDelegate(WriteReport);
            var asyncState = new StateAsync(writeReportDelegate, callback);
            writeReportDelegate.BeginInvoke(report, timeout, EndWriteReport, asyncState);
        }

        /// <summary>
        /// Handle data transfers on the control channel.  This method places data on the control channel for devices
        /// that do not support the interupt transfers
        /// </summary>
        /// <param name="report">The outbound HID report</param>
        /// <returns>The result of the tranfer request: true if successful otherwise false</returns>
        /// 
        public bool WriteReportSync(HidReport report)
        {

            if (null != report)
            {
                byte[] buffer = report.GetBytes();
                return (NativeMethods.HidD_SetOutputReport(WriteHandle, buffer, buffer.Length));
            }
            else
            {
                throw new ArgumentException("The output report is null, it must be allocated before you call this method", "report");
            }
        }

        public async Task<bool> WriteReportAsync(HidReport report, int timeout = 0)
        {
            var writeReportDelegate = new WriteReportDelegate(WriteReport);
#if NET20 || NET35 || NET5_0_OR_GREATER
            return await Task<bool>.Factory.StartNew(() => writeReportDelegate.Invoke(report, timeout));
#else
            return await Task<bool>.Factory.FromAsync(writeReportDelegate.BeginInvoke, writeReportDelegate.EndInvoke, report, timeout, null);
#endif
        }

        public HidReport CreateReport()
        {
            return new HidReport(Capabilities.OutputReportByteLength);
        }

        public bool WriteFeatureData(byte[] data)
        {
            if (_deviceCapabilities.FeatureReportByteLength <= 0)
            {
                return false;
            }

            var buffer = CreateFeatureOutputBuffer();

            Array.Copy(data, 0, buffer, 0, Math.Min(data.Length, _deviceCapabilities.FeatureReportByteLength));


            IntPtr hidHandle = IntPtr.Zero;
            bool success = false;
            try
            {
                if (IsOpen)
                {
                    hidHandle = WriteHandle;
                }
                else
                {
                    hidHandle = OpenDeviceIO(_devicePath, NativeMethods.ACCESS_NONE);
                }

                //var overlapped = new NativeOverlapped();
                success = NativeMethods.HidD_SetFeature(hidHandle, buffer, buffer.Length);
            }
            catch (Exception exception)
            {
                throw new Exception(string.Format("EonaCat HID: Error accessing HID device '{0}'.", _devicePath), exception);
            }
            finally
            {
                if (hidHandle != IntPtr.Zero && hidHandle != WriteHandle)
                {
                    CloseDeviceIO(hidHandle);
                }
            }
            return success;
        }

        protected static void EndRead(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (ReadDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (ReadCallback)hidAsyncState.CallbackDelegate;
            var data = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(data);
            }
        }

        protected static void EndReadReport(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (ReadReportDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (ReadReportCallback)hidAsyncState.CallbackDelegate;
            var report = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(report);
            }
        }

        private static void EndWrite(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (WriteDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (WriteCallback)hidAsyncState.CallbackDelegate;
            var result = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(result);
            }
        }

        private static void EndWriteReport(IAsyncResult ar)
        {
            var hidAsyncState = (StateAsync)ar.AsyncState;
            var callerDelegate = (WriteReportDelegate)hidAsyncState.CallerDelegate;
            var callbackDelegate = (WriteCallback)hidAsyncState.CallbackDelegate;
            var result = callerDelegate.EndInvoke(ar);

            if ((callbackDelegate != null))
            {
                callbackDelegate.Invoke(result);
            }
        }

        private byte[] CreateInputBuffer()
        {
            return CreateBuffer(Capabilities.InputReportByteLength - 1);
        }

        private byte[] CreateOutputBuffer()
        {
            return CreateBuffer(Capabilities.OutputReportByteLength - 1);
        }

        private byte[] CreateFeatureOutputBuffer()
        {
            return CreateBuffer(Capabilities.FeatureReportByteLength - 1);
        }

        private static byte[] CreateBuffer(int length)
        {
            byte[] buffer = null;
            Array.Resize(ref buffer, length + 1);
            return buffer;
        }

        private static HidDeviceAttributes GetDeviceAttributes(IntPtr hidHandle)
        {
            var deviceAttributes = default(NativeMethods.HIDD_ATTRIBUTES);
            deviceAttributes.Size = Marshal.SizeOf(deviceAttributes);
            NativeMethods.HidD_GetAttributes(hidHandle, ref deviceAttributes);
            return new HidDeviceAttributes(deviceAttributes);
        }

        private static HidDeviceCapabilities GetDeviceCapabilities(IntPtr hidHandle)
        {
            var capabilities = default(NativeMethods.HIDP_CAPS);
            var preparsedDataPointer = default(IntPtr);

            if (NativeMethods.HidD_GetPreparsedData(hidHandle, ref preparsedDataPointer))
            {
                NativeMethods.HidP_GetCaps(preparsedDataPointer, ref capabilities);
                NativeMethods.HidD_FreePreparsedData(preparsedDataPointer);
            }
            return new HidDeviceCapabilities(capabilities);
        }

        private bool WriteData(byte[] data, int timeout)
        {
            if (_deviceCapabilities.OutputReportByteLength <= 0)
            {
                return false;
            }

            var buffer = CreateOutputBuffer();
            uint bytesWritten;

            Array.Copy(data, 0, buffer, 0, Math.Min(data.Length, _deviceCapabilities.OutputReportByteLength));

            if (_deviceWriteMode == DeviceMode.Overlapped)
            {
                var security = new NativeMethods.SECURITY_ATTRIBUTES();
                var overlapped = new NativeOverlapped();

                var overlapTimeout = timeout <= 0 ? NativeMethods.WAIT_INFINITE : timeout;

                security.lpSecurityDescriptor = IntPtr.Zero;
                security.bInheritHandle = true;
                security.nLength = Marshal.SizeOf(security);

                overlapped.OffsetLow = 0;
                overlapped.OffsetHigh = 0;
                overlapped.EventHandle = NativeMethods.CreateEvent(ref security, Convert.ToInt32(false), Convert.ToInt32(true), "");

                try
                {
                    NativeMethods.WriteFile(WriteHandle, buffer, (uint)buffer.Length, out bytesWritten, ref overlapped);
                }
                catch { return false; }

                var result = NativeMethods.WaitForSingleObject(overlapped.EventHandle, overlapTimeout);

                switch (result)
                {
                    case NativeMethods.WAIT_OBJECT_0:
                        return true;
                    case NativeMethods.WAIT_TIMEOUT:
                        return false;
                    case NativeMethods.WAIT_FAILED:
                        return false;
                    default:
                        return false;
                }
            }
            else
            {
                try
                {
                    var overlapped = new NativeOverlapped();
                    return NativeMethods.WriteFile(WriteHandle, buffer, (uint)buffer.Length, out bytesWritten, ref overlapped);
                }
                catch { return false; }
            }
        }

        protected DeviceData ReadData(int timeout)
        {
            var buffer = new byte[] { };
            var status = DeviceData.ReadStatus.NoDataRead;
            IntPtr nonManagedBuffer;

            if (_deviceCapabilities.InputReportByteLength > 0)
            {
                uint bytesRead;

                buffer = CreateInputBuffer();
                nonManagedBuffer = Marshal.AllocHGlobal(buffer.Length);

                if (_deviceReadMode == DeviceMode.Overlapped)
                {
                    var security = new NativeMethods.SECURITY_ATTRIBUTES();
                    var overlapped = new NativeOverlapped();
                    var overlapTimeout = timeout <= 0 ? NativeMethods.WAIT_INFINITE : timeout;

                    security.lpSecurityDescriptor = IntPtr.Zero;
                    security.bInheritHandle = true;
                    security.nLength = Marshal.SizeOf(security);

                    overlapped.OffsetLow = 0;
                    overlapped.OffsetHigh = 0;
                    overlapped.EventHandle = NativeMethods.CreateEvent(ref security, Convert.ToInt32(false), Convert.ToInt32(true), string.Empty);

                    try
                    {
                        var success = NativeMethods.ReadFile(ReadHandle, nonManagedBuffer, (uint)buffer.Length, out bytesRead, ref overlapped);

                        if (success) 
                        {
                            status = DeviceData.ReadStatus.Success; // No check here to see if bytesRead > 0 . Perhaps not necessary?
                        }
                        else
                        {
                            var result = NativeMethods.WaitForSingleObject(overlapped.EventHandle, overlapTimeout);

                            switch (result) 
                            {
                                case NativeMethods.WAIT_OBJECT_0:
                                    status = DeviceData.ReadStatus.Success;
                                    NativeMethods.GetOverlappedResult(ReadHandle, ref overlapped, out bytesRead, false);
                                    break;
                                case NativeMethods.WAIT_TIMEOUT:
                                    status = DeviceData.ReadStatus.WaitTimedOut;
                                    buffer = new byte[] { };
                                    break;
                            case NativeMethods.WAIT_FAILED:
                                    status = DeviceData.ReadStatus.WaitFail;
                                    buffer = new byte[] { };
                                    break;
                                default:
                                    status = DeviceData.ReadStatus.NoDataRead;
                                    buffer = new byte[] { };
                                    break;
                            }
                        }
                        Marshal.Copy(nonManagedBuffer, buffer, 0, (int)bytesRead);
                    }
                    catch { status = DeviceData.ReadStatus.ReadError; }
                    finally {
                        CloseDeviceIO(overlapped.EventHandle);
                        Marshal.FreeHGlobal(nonManagedBuffer);
                    }
                }
                else
                {
                    try
                    {
                        var overlapped = new NativeOverlapped();

                        NativeMethods.ReadFile(ReadHandle, nonManagedBuffer, (uint)buffer.Length, out bytesRead, ref overlapped);
                        status = DeviceData.ReadStatus.Success;
                        Marshal.Copy(nonManagedBuffer, buffer, 0, (int)bytesRead);
                    }
                    catch { status = DeviceData.ReadStatus.ReadError; }
                    finally { Marshal.FreeHGlobal(nonManagedBuffer); }
                }
            }
            return new DeviceData(buffer, status);
        }

        private static IntPtr OpenDeviceIO(string devicePath, uint deviceAccess)
        {
            return OpenDeviceIO(devicePath, DeviceMode.NonOverlapped, deviceAccess, ShareMode.ShareRead | ShareMode.ShareWrite);
        }

        private static IntPtr OpenDeviceIO(string devicePath, DeviceMode deviceMode, uint deviceAccess, ShareMode shareMode)
        {
            var security = new NativeMethods.SECURITY_ATTRIBUTES();
            var flags = 0;

            if (deviceMode == DeviceMode.Overlapped)
            {
                flags = NativeMethods.FILE_FLAG_OVERLAPPED;
            }

            security.lpSecurityDescriptor = IntPtr.Zero;
            security.bInheritHandle = true;
            security.nLength = Marshal.SizeOf(security);

            return NativeMethods.CreateFile(devicePath, deviceAccess, (int)shareMode, ref security, NativeMethods.OPEN_EXISTING, flags, hTemplateFile: IntPtr.Zero);
        }

        private static void CloseDeviceIO(IntPtr handle)
        {
            if (Environment.OSVersion.Version.Major > 5)
            {
                NativeMethods.CancelIoEx(handle, IntPtr.Zero);
            }
            NativeMethods.CloseHandle(handle);
        }

        private void DeviceEventMonitorInserted()
        {
            if (!IsOpen)
            {
                OpenDevice(_deviceReadMode, _deviceWriteMode, _deviceShareMode);
            }

            Inserted?.Invoke();
        }

        private void DeviceEventMonitorRemoved()
        {
            if (IsOpen)
            {
                CloseDevice();
            }

            Removed?.Invoke();
        }

        public void Dispose()
        {
            if (MonitorDeviceEvents)
            {
                MonitorDeviceEvents = false;
            }

            if (IsOpen)
            {
                CloseDevice();
            }
        }
    }
}
